unit zBalloon;

interface

uses
  Windows, SysUtils, Classes, Forms, ExtCtrls, Controls, Messages, Graphics,
  StdCtrls, ImgList, ComCtrls, MMSystem;

type
  TIconMode = (imNone, imGlyph, imImageList, imApplIcon);
  TShowMode = (smDefault, smHint, smMessage, smError, smWorning, smInput, smDialog);
  TInputType = (itString, itInteger, itDouble, itPassword, itFileName, itMail,
                itURL, itCheckBox, itRadioButton, itCombo, itComboList);
  TButtonType = (btYN, btOC, btYNC, btOK);

  TzBalloon = class;

  TInputEdit = class(TPersistent)
  private
    { Private declarations }
    FFont: TFont;
    FColor: TColor;
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TzBalloon);  virtual;
    destructor  Destroy; override;
  published
    { Published declarations }
    property Font : TFont read FFont write FFont;
    property Color: TColor read FColor write FColor default clInfoBk;
  end;

  TCustomOptions = class(TPersistent)
  private
    { Private declarations }
    FFont: TFont;
    FTitleFont: TFont;
    FWallPaper: TBitmap;
    FTopColor: TColor;
    FBottomColor: TColor;
    procedure SetWallPaper(Value: TBitmap);
    function GetWallPaper: TBitmap;
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TzBalloon);  virtual;
    destructor  Destroy; override;
    property Font : TFont read FFont write FFont;
    property TitleFont: TFont read FTitleFont write FTitleFont;
    property WallPaper: TBitmap read GetWallPaper write SetWallPaper;
    property TopColor: TColor read FTopColor write FTopColor default $00FDFDFD;
    property BottomColor: TColor read FBottomColor write FBottomColor default clSilver;
  published
    { Published declarations }
  end;

  TButtonOptions = class(TCustomOptions)
  published
    property Font;
    property WallPaper;
    property TopColor;
    property BottomColor;
  end;

  THintOptions = class(TCustomOptions)
  published
    property Font;
    property WallPaper;
    property TopColor;
    property BottomColor;
    property TitleFont;
  end;

{  TMsgOptions = class(TCustomOptions)
  published
    property Font;
    property WallPaper;
    property TopColor;
    property BottomColor;
    property TitleFont;
  end;}

  TInputOptions = class(TCustomOptions)
  private
    FEdit: TInputEdit;
    FMin,
    FMax: Double;
    FDecimalCount: Byte;
    FCheckBoxCaption: string;
    FCheckBoxChecked: Boolean;
    FButtonOptions: TButtonOptions;
    fBoxList: TStrings;
    fRadioIndex: Integer;
    fComboList: TStrings;

    procedure SetBoxList(Value: TStrings);
    procedure SetComboList(Value: TStrings);
  public
    constructor Create(AOwner: TzBalloon);  override;
    destructor Destroy; override;
  published
    property Font;
    property WallPaper;
    property TopColor;
    property BottomColor;
    property TitleFont;
    property Edit: TInputEdit read FEdit write FEdit;
    property Min: Double read FMin write FMin;
    property Max: Double read FMax write FMax;
    property DecimalCount: Byte read FDecimalCount write FDecimalCount default 2;
    property CheckBoxCaption: string read FCheckBoxCaption write FCheckBoxCaption;
    property CheckBoxChecked: Boolean read FCheckBoxChecked write FCheckBoxChecked Default False;
    property ButtonOptions: TButtonOptions read FButtonOptions write FButtonOptions;
    property BoxList: TStrings read fBoxList write SetBoxList;
    property ComboList: TStrings read fComboList write SetComboList;
  end;

  TDlgOptions = class(TCustomOptions)
  private
    FCheckBoxCaption: string;
    FCheckBoxChecked: Boolean;
    FButtonOptions: TButtonOptions;
  public
    constructor Create(AOwner: TzBalloon);  override;
  published
    property Font;
    property WallPaper;
    property TopColor;
    property BottomColor;
    property TitleFont;
    property CheckBoxCaption: string read FCheckBoxCaption write FCheckBoxCaption;
    property CheckBoxChecked: Boolean read FCheckBoxChecked write FCheckBoxChecked Default False;
    property ButtonOptions: TButtonOptions read FButtonOptions write FButtonOptions;
  end;

  TButtonCaptions = class(TPersistent)
  private
    FYes,
    FNo,
    FCancel,
    FOk: string;
  public
    constructor Create(AOwner: TzBalloon);
  published
    property Yes: string read FYes write FYes;
    property No: string read FNo write FNo;
    property Cancel: string read FCancel write FCancel;
    property Ok: string read FOk write FOk;  // Thanks to Jean-Christophe Lavergne
  end;

  TzBalloon = class(TComponent)
  private
    { Private declarations }
		FTitle: string;
    FText: TStrings;
    FShowArrow: Boolean;
    FBusy: Boolean;
    FMaxWidth: Word;
    FGlyph: TBitmap;
    FIconMode: TIconMode;
    FImageList: TCustomImageList;
    FImageIndex: TImageIndex;
    FImageWidth: Word;
    FImageHeight: Word;
    FImageStretch: Boolean;
    FShowOnF1: Boolean;
    FOnClick: TNotifyEvent;
    FClickHereText: string;
    FOldOnHint: TShowHintEvent;
    FOldHintMode: Boolean;
    FHintMode: Boolean;
    FHintControl: TControl;
    FOnHint: TShowHintEvent;
    FOnShow: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FShowMode: TShowMode;
    FInputResult: TModalResult;
    FInputValue: Variant;
    FKeyPressEvent: TKeyPressEvent;

    FHintOptions: THintOptions;
    FOptions: THintOptions;
    FMsgOptions: THintOptions;
    FInputOptions: TInputOptions;
    FDlgOptions: TDlgOptions;
    FButtonCaptions: TButtonCaptions;
    FSoundsOff: Boolean;
    FShowHint: Boolean;

    procedure SetGlyph(Value: TBitmap);
    procedure SetImageList(Value: TCustomImageList);
    function GetBitmap: TBitmap;
    procedure SetText(Value: TStrings);
//    procedure SetShowOnF1(Value: Boolean);
    procedure SetHintMode(Value: Boolean);
    procedure DisplayHint(Sender: TObject);
    procedure DoShowHint(var HintStr: String; var CanShow: Boolean; var HintInfo: THintInfo);
    procedure _ShowMsg(Msg: string; TitleText: string; Control: TControl; MsgType: TShowMode);
    function _MsgDlg(ButType: TButtonType; Msg: string; TitleText: string): TModalResult;
    function _Input(InpType: TInputType; Value: Variant; Msg: string; TitleText: string): TModalResult;
    function _InputCombo(var Value: string; ComboType: TInputType; Msg: string;
                         TitleText: string; DefIndex: Integer): Boolean;
  protected
    { Protected declarations }
    FButtonType: TButtonType;
    FInputType: TInputType;
    FButtonOptions: TButtonOptions;
  public
    { Public declarations }
    Constructor Create(AOwner: TComponent); override;
    Destructor Destroy; override;

    procedure Show(X, Y: integer); overload;
    procedure Show(Control: TControl; TmpShowMode: TShowMode = smDefault); overload;
    procedure ShowHint(Control: TControl; NewTitle: string = '');
    procedure ShowMsg(Msg: string; TitleText: string = ''; Control: TControl = nil);
    procedure ShowError(Msg: string; TitleText: string = ''; Control: TControl = nil);
    procedure ShowWorning(Msg: string; TitleText: string = ''; Control: TControl = nil);
    procedure MsgOK(Msg: string; TitleText: string = '');
    function MsgYN(Msg: string; TitleText: string = ''): Boolean;
    function MsgOC(Msg: string; TitleText: string = ''): Boolean;
    function MsgYNC(Msg: string; TitleText: string = ''): TModalResult;
    function Input(var Value: string; Msg: string; TitleText: string = '';
                   KeyPressEvent: TKeyPressEvent = nil): Boolean; overload;
    function Input(var Value: Integer; Msg: string; TitleText: string = ''): Boolean; overload;
    function Input(var Value: Double; Msg: string; TitleText: string = ''): Boolean; overload;
    function InputPassword(var Value: string; Msg: string; TitleText: string = ''): Boolean;
    function InputFileName(var Value: string; Msg: string; TitleText: string = ''): Boolean;
    function InputMail(var Value: string; Msg: string; TitleText: string = ''): Boolean;
    function InputURL(var Value: string; Msg: string; TitleText: string = ''): Boolean;

    function InputCheck(Msg: string; TitleText: string = ''): Boolean;
    function InputRadio(Msg: string; TitleText: string = ''): Integer;

    function InputCombo(var Value: string; Msg: string; TitleText: string = '';
                        DefIndex: Integer = -1): Boolean;
    function InputComboList(var Value: string; Msg: string; TitleText: string = '';
                            DefIndex: Integer = -1): Boolean;

    procedure Close;

    property Busy: Boolean read FBusy write FBusy;
  published
    { Published declarations }
    property Glyph: TBitmap read GetBitmap write SetGlyph;
    property MaxWidth: Word read FMaxWidth write FMaxWidth;
		property Text: TStrings read FText write SetText;
		property Title: String read FTitle write FTitle;
    property IconMode: TIconMode read FIconMode write FIconMode default imApplIcon;
    property ShowArrow: Boolean read FShowArrow write FShowArrow default True;
    property ImageList: TCustomImageList read FImageList write SetImageList;
    property ImageIndex: TImageIndex read FImageIndex write FImageIndex default -1;
    property ClickHereText: string read FClickHereText write FClickHereText;
    property HintMode: Boolean read FHintMode write SetHintMode default False;
    property ImageWidth: Word read FImageWidth write FImageWidth default 16;
    property ImageHeight: Word read FImageHeight write FImageHeight default 16;
    property ImageStretch: Boolean read FImageStretch write FImageStretch default false;
    property SoundsOff: Boolean read FSoundsOff write FSoundsOff default false;

    property HintOptions: THintOptions read FHintOptions write FHintOptions;
    property Options: THintOptions read FOptions write FOptions;
    property MsgOptions: THintOptions read FMsgOptions write FMsgOptions;
    property InputOptions: TInputOptions read FInputOptions write FInputOptions;
    property DlgOptions: TDlgOptions read FDlgOptions write FDlgOptions;
    property ButtonCaptions: TButtonCaptions read FButtonCaptions write FButtonCaptions;
    property HintShow: Boolean read FShowHint write FShowHint default true;

    //property ShowOnF1: Boolean read FShowOnF1 write SetShowOnF1 Default false;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property OnHint: TShowHintEvent read FOnHint write FOnHint;
    property OnShow: TNotifyEvent read FOnShow write FOnShow;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnKeyPress: TKeyPressEvent read FKeyPressEvent write FKeyPressEvent;
  end;

  TzzHtmlLabel = class(TCustomLabel)
  private
    { Private declarations }
    NeedHeight: Integer;
    FCaption: string;
    FDrawing: Boolean;
    FBox: TImage;
    FChecked: Boolean;

    procedure SetCaption(Value: string);
    procedure SetChecked(Value: Boolean);
  protected
    procedure DrawStringEx;
  public
    MaxLineLength: Integer;

    constructor Create(AOwner: TComponent); override;
    procedure Paint; override;
    property Checked: Boolean read FChecked write SetChecked;
  published
    property Caption: string read FCaption write SetCaption;
//    property OnClick;
  end;

  TzzButton = class(TCustomControl)
  private
    FActive: Boolean;
//    FDefault: Boolean;
    FMouseIn: Boolean;
    FDown: Boolean;
    FModalResult: TModalResult;
    FCancel: Boolean;
    procedure CMDialogKey(var Message: TCMDialogKey); message CM_DIALOGKEY;
    procedure CMFocusChanged(var Message: TCMFocusChanged); message CM_FOCUSCHANGED;
    procedure CMMOUSEENTER(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMOUSELEAVE(var Message: TMessage); message CM_MOUSELEAVE;
    procedure WMLBUTTONDOWN( Var msg: TWMLBUTTONDOWN ); message WM_LBUTTONDOWN;
    procedure WMLBUTTONUP( Var msg: TWMLBUTTONUP); message WM_LBUTTONUP;
  protected
//    procedure CreateParams(var Params: TCreateParams); override;
//    procedure CreateWnd; override;
    procedure Paint; override;
    procedure Click; override;
    //procedure WMEraseBkGnd( Var msg: TWMEraseBkGnd ); message WM_ERASEBKGND;
   public
    constructor Create(AOwner: TComponent); override;
  published
    property Cancel: Boolean read FCancel write FCancel default False;
    property ModalResult: TModalResult read FModalResult write FModalResult default 0;
  end;

  TzBalloonCheckBox = class(TCheckBox)
  protected
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure SetButtonStyle;
  end;

  TBalloon = Class(TCustomForm)
  private
    lblTitle: TzzHtmlLabel;
    lblText: TzzHtmlLabel;
    lblClickHere: TLabel;
    iconBitmap: TImage;
    pnlAlign: TPanel;
    tmrExit: TTimer;
    OldOnActiveControlChange: TNotifyEvent;
    OldOnActiveFormChange: TNotifyEvent;
    FEdit: TEdit;
    FComboBox: TComboBox;
    FRadioChecked: TzzHtmlLabel;
    procedure FormPaint(Sender: TObject);
    procedure OnTimer(sender : TObject);
    procedure OnActiveControlChange(sender : TObject);
    procedure OnActiveFormChange(sender : TObject);
    procedure OnYesClick(sender : TObject);
    procedure OnCheckBoxClick(sender : TObject);
    procedure WMNCHitTest(var M: TWMNCHitTest); message wm_NCHitTest;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
//    procedure OnMouseClick(Sender: TObject);
  protected
    Btn1, Btn2, Btn3: TzzButton;
    FCheckBox: TzBalloonCheckBox;
    Procedure CreateParams(Var Params: TCreateParams); override;
    Procedure OnChange(Sender: TObject);
    Procedure OnEditChange(Sender: TObject);
    Procedure OnEditKeyPress(Sender: TObject; var Key: Char);
    Procedure WndProc(Var Message: TMessage); override;
    Procedure OnFormDestroy(Sender: TObject);
  public
    constructor CreateNew(AOwner: TComponent; Dummy: Integer = 0); override;
    procedure InitializeBalloon(X, Y: Integer; ArrowShow: Boolean);
  End;


procedure Register;

const
  cIllegalChars = '\/:*?<>|,' + #34;   // Filename input mode

var F1KeyboardHookHandle: HHook = 0;
    BalloonKeyboardHookHandle, BalloonMouseHookHandle: HHook;
    zBalloonControl: TzBalloon;
    BalloonForm: TBalloon;

implementation
{$R *.RES}
//{$R SOUNDS.RES}

procedure DrawGradient(Wnd : THandle; aTopColor, aEndColor: TColor);
  var
  StartC, EndC: TColor;
  GradRECT: TRect;
  StartRGB, EndRGB: array[0..2] of Byte;
  Colors, i, Delta: Word;
  hdc_ : HDC;
  hbrush_ : HBRUSH;
begin

  hdc_ := GetDC(Wnd);
  if hdc_ = 0 then exit;
  Windows.GetClientRect(Wnd, GradRect);

  StartC := ColorToRGB(aTopColor);
  EndC := ColorToRGB(aEndColor);

  StartRGB[0] := GetRValue(StartC);
  StartRGB[1] := GetGValue(StartC);
  StartRGB[2] := GetBValue(StartC);

  EndRGB[0] := GetRValue(EndC);
  EndRGB[1] := GetGValue(EndC);
  EndRGB[2] := GetBValue(EndC);

  Colors := ((GradRect.Bottom - GradRect.Top) div 2) ;
  Delta := (GradRect.Bottom - GradRect.Top) div Colors;

   for i := 0 to Colors - 1 do begin
    GradRECT.Top := i*Delta;
    GradRECT.Bottom := GradRECT.Top + Delta;
    hBrush_ := CreateSolidBrush(RGB(
    (StartRGB[0] + MulDiv(i, EndRGB[0] - StartRGB[0], Colors-1)),
    (StartRGB[1] + MulDiv(i, EndRGB[1] - StartRGB[1], Colors-1)),
    (StartRGB[2] + MulDiv(i, EndRGB[2] - StartRGB[2], Colors-1))));
     FillRect(hDC_, GradRECT, hBrush_);
     DeleteObject(hBrush_);
   end;
//   ReleaseDC(IssOptionsF.Handle, hDC_);
end;

function F1_Key_HooK(code: Integer; wparam: WPARAM; lparam: LPARAM): LRESULT stdcall;
begin
 if (Code = 0) and ((lParam and $40000000) = 0) then begin
   if not zBalloonControl.FBusy and
      (GetKeyState(VK_SHIFT) >=0) and
      (GetKeyState(VK_CONTROL) >= 0) and
      (GetKeyState(VK_MENU) >= 0) then
     zBalloonControl.Show(Screen.ActiveControl);
   Result := 0;
 end
 else Result := CallNextHookEx(BalloonKeyboardHookHandle, Code, wParam, lParam);
end;

constructor TzBalloon.Create(AOwner: TComponent);
Begin
	Inherited;
  zBalloonControl := Self;
	FText := TStringList.Create;
  FBusy := False;
  FMaxWidth := Screen.Width div 4;
  FShowArrow := True;
  FIconMode := imApplIcon;
  FImageList := nil;
  FImageIndex := -1;
  FShowOnF1 := False;
  F1KeyboardHookHandle := 0;
  FClickHereText := 'Click here or press ENTER for more information...';
  FHintControl := nil;
  FImageWidth := 16;
  FImageHeight := 16;
  FImageStretch := False;
  FShowMode := smDefault;
  FSoundsOff := False;
  FShowHint := True;

  FHintOptions := THintOptions.Create(Self);
  FOptions := THintOptions.Create(Self);
  FMsgOptions := THintOptions.Create(Self);
  FInputOptions := TInputOptions.Create(Self);
  FDlgOptions := TDlgOptions.Create(Self);
  FButtonCaptions := TButtonCaptions.Create(Self);

end;

destructor TzBalloon.Destroy;
Begin
  FText.Free;
  FHintOptions.Free;
  if Assigned(FGlyph) then FGlyph.Free;
  if F1KeyboardHookHandle <> 0 then UnhookWindowsHookEx(F1KeyboardHookHandle);
	Inherited;
End;

procedure TzBalloon.ShowHint(Control: TControl; NewTitle: string = '');
begin                    
//  if not (Control is TWinControl) then Exit;
  if NewTitle <> '' then FTitle := NewTitle;
//  FText.Text := (Control as TWinControl).Hint;
  FText.Text := Control.Hint;
  if Length(Text.Text) > Screen.Width div 4 then
    FMaxWidth := Screen.Width div 4
  else FMaxWidth := Screen.Width div 8;
  Show(Control);
end;

procedure TzBalloon.Show(Control: TControl; TmpShowMode: TShowMode = smDefault);
var FControlRect: TRect;
begin
  if not Assigned(Control) then Exit;
  if (Control is TWinControl) then
    GetWindowRect((Control as TWinControl).Handle, FControlRect)
  else begin
    GetWindowRect((Control.Parent as TWinControl).Handle, FControlRect);
    with FControlRect do begin
      Left := Left + Control.Left + GetSystemMetrics(SM_CXFIXEDFRAME);
      Top := Top + Control.Top +
             GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYFIXEDFRAME);
      Right := Left + Control.Width;
      Bottom := Top + Control.Height;
    end;
  end;
  FShowMode := TmpShowMode;

  Close;
  FShowArrow := True;
  if (FShowMode <> smHint) and Assigned(FOnShow) then FOnShow(Self);
	BalloonForm := TBalloon.CreateNew(Owner);
  FBusy := True;
	BalloonForm.InitializeBalloon(FControlRect.Left + (FControlRect.Right - FControlRect.Left) div 2,
       FControlRect.Top + (FControlRect.Bottom - FControlRect.Top) div 2,
       not (Control is TCustomForm) and FShowArrow);
end;

procedure TzBalloon.Show(X, Y: integer);
begin
  Close;
  FShowArrow := not (FShowMode in [smMessage, smError, smWorning, smInput, smDialog]);
  if (FShowMode <> smHint) and Assigned(FOnShow) then FOnShow(Self);
	BalloonForm := TBalloon.CreateNew(Owner);
  FBusy := True;
	BalloonForm.InitializeBalloon(X, Y, FShowArrow);
end;

procedure TzBalloon.Close;
begin
  if Assigned(BalloonForm) then begin
    BalloonForm.Release;
    Application.ProcessMessages;
    BalloonForm := nil;
    FBusy := False;
  end;
end;

procedure TzBalloon.SetText(Value: TStrings);
begin
  FText.Assign(Value);
end;

procedure TzBalloon.SetGlyph(Value: TBitmap);
begin
  if FGlyph = nil then begin
    FGlyph := TBitmap.Create;
    FGlyph.Transparent := True;
  end;
  FGlyph.Assign(Value);
  if not (csLoading in ComponentState) then begin
    FImageWidth := FGlyph.Width;
    FImageHeight := FGlyph.Height;
  end;
end;

procedure TzBalloon.SetImageList(Value: TCustomImageList);
begin
  if FImageList <> Value then begin
    FImageList := Value;
    if not (csLoading in ComponentState) then begin
      FImageWidth := FImageList.Width;
      FImageHeight := FImageList.Height;
    end;
  end;
end;

function TzBalloon.GetBitmap: TBitmap;
begin
  if FGlyph = nil then begin
    FGlyph := TBitmap.Create;
    FGlyph.Transparent := True;
  end;
  Result := FGlyph;
end;

{procedure TzBalloon.SetShowOnF1(Value: Boolean);
begin
  if FShowOnF1 = Value then Exit;
  FShowOnF1 := Value;
  if not (CSDesigning in ComponentState) then begin
    if FShowOnF1 then
      F1KeyboardHookHandle := SetWindowsHookEx(WH_KEYBOARD, F1_Key_HooK, hInstance, GetCurrentThreadID)
    else if F1KeyboardHookHandle <> 0 then UnhookWindowsHookEx(F1KeyboardHookHandle);
  end;
end;}

procedure TzBalloon.SetHintMode(Value: Boolean);
begin
  if FHintMode <> Value then begin
    FHintMode := Value;
    if csDesigning in ComponentState then Exit;
    if FHintMode then begin
      FOldHintMode := Application.ShowHint;
//      Application.ShowHint := False;
      FOldOnHint := Application.OnShowHint; //Application.OnHint;
      Application.OnShowHint := DoShowHint;
      Application.OnHint := DisplayHint;
    end
    else begin
      Application.ShowHint := FOldHintMode;
      Application.OnShowHint := FOldOnHint;
      Application.OnHint := nil;
    end;
  end;
end;

procedure TzBalloon.DoShowHint(var HintStr: string; var CanShow: Boolean; var HintInfo: THintInfo);
var s: string;
    i: Integer;
begin
  if Busy or not FShowHint or (FHintControl = HintInfo.HintControl) or (HintStr = '') then begin
    CanShow := False;
    Exit;
  end;
  s := HintInfo.HintControl.Hint;
  i := Pos('|', s);
  if i > 0 then begin
    FTitle := Copy(s, 1, i - 1);
    s := Copy(s, i + 1, MaxInt);
  end
  else FTitle := Application.Title;

  if (HintInfo.HintControl is TToolButton) and
     Assigned(ImageList) and
     (((HintInfo.HintControl as TToolButton).Parent as TToolBar).Images = ImageList) then begin
    if i = 0 then
      Title := (HintInfo.HintControl as TToolButton).Caption;
    ImageIndex := (HintInfo.HintControl as TToolButton).ImageIndex;
    IconMode := imImageList;
  end
  else if Assigned(FGlyph) then
    IconMode := imGlyph
  else IconMode := imApplIcon;

  if Assigned(FOnHint) then
    FOnHint(s, CanShow, HintInfo);
  if not CanShow then Exit;
  
  FText.Text := s;
  CanShow := False;
  FShowMode := smHint;
  FShowArrow := True;
  if Length(Text.Text) > Screen.Width div 4 then
    FMaxWidth := Screen.Width div 4
  else FMaxWidth := Screen.Width div 8;
  FHintControl := HintInfo.HintControl;
  Show(Mouse.CursorPos.X, Mouse.CursorPos.Y);
  FMaxWidth := Screen.Width div 4;
end;

procedure TzBalloon.DisplayHint(Sender: TObject);
begin
  if (FShowMode = smHint) then
    Close;
  FHintControl := nil;
end;

procedure TzBalloon._ShowMsg(Msg: string; TitleText: string; Control: TControl; MsgType: TShowMode);
begin
  FShowMode := MsgType;
  FShowArrow := Assigned(Control);
  FText.Text := Msg;
  Title := TitleText;
  if not Assigned(Control) and Assigned(Screen.ActiveCustomForm) then begin
    with Screen.ActiveCustomForm do
      Self.Show(Left + Width div 2, Top + Height div 2)
  end
  else Show(Control, MsgType);
end;

procedure TzBalloon.ShowMsg(Msg: string; TitleText: string = ''; Control: TControl = nil);
begin
  if not FSoundsOff then begin
    SystemParametersInfo(SPI_SETBEEP, 1, nil, 0); //ON
    MessageBeep(MB_OK);
  end;
  _ShowMsg(Msg, TitleText, Control, smMessage);
end;

procedure TzBalloon.ShowError(Msg: string; TitleText: string = ''; Control: TControl = nil);
begin
  if not FSoundsOff then begin
    SystemParametersInfo(SPI_SETBEEP, 1, nil, 0); //ON
    MessageBeep(MB_ICONSTOP);
  end;
  _ShowMsg(Msg, TitleText, Control, smError);
end;

procedure TzBalloon.ShowWorning(Msg: string; TitleText: string = ''; Control: TControl = nil);
begin
  if not FSoundsOff then begin
    SystemParametersInfo(SPI_SETBEEP, 1, nil, 0); //ON
    MessageBeep(MB_ICONASTERISK);
  end;
  _ShowMsg(Msg, TitleText, Control, smWorning);
end;

function TzBalloon._MsgDlg(ButType: TButtonType; Msg: string; TitleText: string): TModalResult;
begin
  FShowMode := smDialog;
  FText.Text := Msg;
  Title := TitleText;
  FShowArrow := False;
  FButtonType := ButType;
  if not FSoundsOff then begin
    SystemParametersInfo(SPI_SETBEEP, 1, nil, 0); //ON
    MessageBeep(MB_OK);
  end;
  if Assigned(Screen.ActiveCustomForm) then
    with Screen.ActiveCustomForm do
      Self.Show(Left + Width div 2, Top + Height div 2)
  else with Screen do
    Self.Show(Width div 2, Height div 2);
  Result := FInputResult;
  Close;
end;

procedure TzBalloon.MsgOK(Msg: string; TitleText: string = '');
begin
  _MsgDlg(btOK, Msg, TitleText);
end;

function TzBalloon.MsgYN(Msg: string; TitleText: string = ''): Boolean;
begin
  Result := _MsgDlg(btYN, Msg, TitleText) = mrYes;
end;

function TzBalloon.MsgOC(Msg: string; TitleText: string = ''): Boolean;
begin
  Result := _MsgDlg(btOC, Msg, TitleText) = mrYes;
end;

function TzBalloon.MsgYNC(Msg: string; TitleText: string = ''): TModalResult;
begin
  Result := _MsgDlg(btYNC, Msg, TitleText);
end;

function TzBalloon._Input(InpType: TInputType; Value: Variant; Msg: string; TitleText: string): TModalResult;
begin
  FButtonType := btOC;
  FShowMode := smInput;
  FText.Text := Msg;
  Title := TitleText;
  FShowArrow := False;
  FInputValue := Value;
  FInputType := InpType;
  if not FSoundsOff then begin
    SystemParametersInfo(SPI_SETBEEP, 1, nil, 0); //ON
    PlaySound(PChar('INPUT_SOUND'), 0, SND_RESOURCE or SND_ASYNC);
  end;
  with Screen.ActiveCustomForm do
    Self.Show(Left + Width div 2, Top + Height div 2);
  Result := FInputResult;
  Close;
end;

function TzBalloon.Input(var Value: string; Msg: string; TitleText: string = '';
                         KeyPressEvent: TKeyPressEvent = nil): Boolean;
begin
  Result := _Input(itString, Value, Msg, TitleText) = mrYes;
  if Result then Value := FInputValue;
end;

function TzBalloon.Input(var Value: Integer; Msg: string; TitleText: string = ''): Boolean;
begin
  Result := _Input(itInteger, Value, Msg, TitleText) = mrYes;
  if Result then Value := FInputValue;
end;

function TzBalloon.Input(var Value: Double; Msg: string; TitleText: string = ''): Boolean;
begin
  Result := _Input(itDouble, Value, Msg, TitleText) = mrYes;
  if Result then Value := FInputValue;
end;

function TzBalloon.InputPassword(var Value: string; Msg: string; TitleText: string = ''): Boolean;
begin
  Result := _Input(itPassword, Value, Msg, TitleText) = mrYes;
  if Result then Value := FInputValue;
end;

function TzBalloon.InputFileName(var Value: string; Msg: string; TitleText: string = ''): Boolean;
begin
  FInputOptions.FMin := 1;
//  FInputOptions.FMax := PathMax;
  Result := _Input(itFileName, Value, Msg, TitleText) = mrYes;
  if Result then Value := FInputValue;
end;

function TzBalloon.InputMail(var Value: string; Msg: string; TitleText: string = ''): Boolean;
begin
  FInputOptions.FMin := 1;
  Result := _Input(itMail, Value, Msg, TitleText) = mrYes;
  if Result then Value := FInputValue;
end;

function TzBalloon.InputURL(var Value: string; Msg: string; TitleText: string = ''): Boolean;
begin
  FInputOptions.FMin := 4;
  if Value = '' then Value := 'http://';
  Result := _Input(itURL, Value, Msg, TitleText) = mrYes;
  if Result then Value := FInputValue;
end;

function TzBalloon.InputCheck(Msg: string; TitleText: string = ''): Boolean;
begin
  FButtonType := btOC;
  FShowMode := smInput;
  FText.Text := Msg;
  Title := TitleText;
  FShowArrow := False;
  FInputType := itCheckBox;
  if not FSoundsOff then begin
    SystemParametersInfo(SPI_SETBEEP, 1, nil, 0); //ON
    PlaySound(PChar('INPUT_SOUND'), 0, SND_RESOURCE or SND_ASYNC);
  end;
  with Screen.ActiveCustomForm do
    Self.Show(Left + Width div 2, Top + Height div 2);
  Result := FInputResult = mrYes;
  Close;
end;

function TzBalloon.InputRadio(Msg: string; TitleText: string = ''): Integer;
begin
  FButtonType := btOC;
  FShowMode := smInput;
  FText.Text := Msg;
  Title := TitleText;
  FShowArrow := False;
  FInputType := itRadioButton;
  InputOptions.fRadioIndex := -1;
  if not FSoundsOff then begin
    SystemParametersInfo(SPI_SETBEEP, 1, nil, 0); //ON
    PlaySound(PChar('INPUT_SOUND'), 0, SND_RESOURCE or SND_ASYNC);
  end;
  with Screen.ActiveCustomForm do
    Self.Show(Left + Width div 2, Top + Height div 2);
  if FInputResult = mrYes then
    Result := InputOptions.fRadioIndex
  else Result := -1;
  Close;
end;

function TzBalloon._InputCombo(var Value: string; ComboType: TInputType; Msg: string; TitleText: string;
                               DefIndex: Integer): Boolean;
begin
  FButtonType := btOC;
  FShowMode := smInput;
  FText.Text := Msg;
  Title := TitleText;
  FShowArrow := False;
  FInputType := ComboType;
  InputOptions.fRadioIndex := DefIndex;
  if not FSoundsOff then begin
    SystemParametersInfo(SPI_SETBEEP, 1, nil, 0); //ON
    PlaySound(PChar('INPUT_SOUND'), 0, SND_RESOURCE or SND_ASYNC);
  end;
  with Screen.ActiveCustomForm do
    Self.Show(Left + Width div 2, Top + Height div 2);
  if FInputResult = mrYes then begin
    Result := True;
    Value := FInputValue;
  end
  else begin
    Result := False;
  end;
  Close;
end;

function TzBalloon.InputCombo(var Value: string; Msg: string; TitleText: string = '';
                              DefIndex: Integer = -1): Boolean;
begin
  Result := _InputCombo(Value, itCombo, Msg, TitleText, DefIndex);
end;

function TzBalloon.InputComboList(var Value: string; Msg: string; TitleText: string = '';
                                  DefIndex: Integer = -1): Boolean;
begin
  Result := _InputCombo(Value, itComboList, Msg, TitleText, DefIndex);
end;

///////////////////////////////  Balloon form

constructor TBalloon.CreateNew(AOwner: TComponent; Dummy: Integer = 0);
begin
  Inherited;

  OnActivate   := OnChange;
  OnDeactivate := OnChange;
  OnShow       := OnChange;
  BorderStyle  := bsNone;
  FormStyle    := fsStayOnTop;
  OnPaint      := FormPaint;
  FRadioChecked := nil;

  pnlAlign   := TPanel.Create(Self);
  lblTitle   := TzzHtmlLabel.Create(Self);
  lblText    := TzzHtmlLabel.Create(Self);
  lblClickHere := TLabel.Create(Self);
  FCheckBox := nil;

  case zBalloonControl.FShowMode of
    smInput: begin
//               Color := zBalloonControl.FInputOptions.TopColor;
               Font.Assign(zBalloonControl.FInputOptions.FFont);
               lblTitle.Font.Assign(zBalloonControl.FInputOptions.FTitleFont);
             end;
    smHint: begin
//              Color := zBalloonControl.FHintOptions.TopColor;
              Font.Assign(zBalloonControl.FHintOptions.FFont);
              lblTitle.Font.Assign(zBalloonControl.FHintOptions.FTitleFont);
            end;
    smDialog: begin
//                 Color := zBalloonControl.FDlgOptions.TopColor;
                 Font.Assign(zBalloonControl.FDlgOptions.FFont);
                 lblTitle.Font.Assign(zBalloonControl.FDlgOptions.FTitleFont);
               end;
    smMessage,
    smError,
    smWorning: begin
//                 Color := zBalloonControl.FMsgOptions.TopColor;
                 Font.Assign(zBalloonControl.FMsgOptions.FFont);
                 lblTitle.Font.Assign(zBalloonControl.FMsgOptions.FTitleFont);
               end;
    else begin
//      Color := zBalloonControl.FOptions.TopColor;
      Font.Assign(zBalloonControl.FOptions.FFont);
      lblTitle.Font.Assign(zBalloonControl.FOptions.FTitleFont);
    end;
  end;

  iconBitmap := TImage.Create(Self);
  if not (zBalloonControl.FShowMode in [smInput, smDialog]) and zBalloonControl.FImageStretch then begin
    iconBitmap.Height := zBalloonControl.FImageHeight;
    iconBitmap.Width := zBalloonControl.FImageWidth;
    iconBitmap.Stretch := True;
  end
  else iconBitmap.AutoSize := True;

  tmrExit := TTimer.Create(Self);
  tmrExit.Enabled := False;
  tmrExit.Interval := 100;
  tmrExit.OnTimer := OnTimer;

  lblTitle.Parent      := Self;
  lblTitle.ParentColor := True;
  lblTitle.Transparent := True;
//  lblTitle.ParentFont  := True;
  //lblTitle.AutoSize    := True;
//  lblTitle.Font.Style  := [fsBold];
  lblTitle.Left        := 34;
  lblTitle.Top         := 12;

  lblText.Parent      := Self;
  lblText.ParentColor := True;
  lblText.ParentFont  := True;
  lblText.Transparent := True;
  lblText.Left        := 10;

  lblClickHere.Parent := Self;
  lblClickHere.ParentColor := True;
  lblClickHere.Font.Color := clBlue;
  lblClickHere.Font.Style := [fsBold];
  lblClickHere.Cursor := crHandPoint;
  lblClickHere.WordWrap := True;
  lblClickHere.Transparent := True;
  lblClickHere.Visible := (zBalloonControl.FShowMode = smDefault) and
                          Assigned(zBalloonControl.OnClick) and
                          (zBalloonControl.FClickHereText <> '');

  iconBitmap.Parent      := Self;
  iconBitmap.Transparent := True;
  iconBitmap.Left        := 10;
  iconBitmap.Top         := 10;

  OnDestroy := OnFormDestroy;
  BalloonForm := Self;
  BalloonKeyboardHookHandle := 0;
  BalloonMouseHookHandle := 0;
End;

Procedure TBalloon.OnFormDestroy(Sender: TObject);
begin
  if BalloonKeyboardHookHandle <> 0 then
    UnhookWindowsHookEx(BalloonKeyboardHookHandle);
  if BalloonMouseHookHandle <> 0 then
    UnhookWindowsHookEx(BalloonMouseHookHandle);
  BalloonKeyboardHookHandle := 0;
  BalloonMouseHookHandle := 0;
  Screen.OnActiveControlChange := OldOnActiveControlChange;
  Screen.OnActiveFormChange := OldOnActiveFormChange;
  BalloonForm := nil;
  tmrExit.Free;
  zBalloonControl.FBusy := False;
  zBalloonControl.FInputOptions.FMin := 0;
  zBalloonControl.FInputOptions.FMax := 0;
  zBalloonControl.FDlgOptions.FCheckBoxCaption := '';
  zBalloonControl.FInputOptions.FCheckBoxCaption := '';
  if zBalloonControl.FShowMode = smInput then begin
    if Assigned(FCheckBox) then
      zBalloonControl.FInputOptions.FCheckBoxChecked := FCheckBox.Checked;
  end
  else if zBalloonControl.FShowMode = smDialog then begin
    if Assigned(FCheckBox) then
      zBalloonControl.FDlgOptions.FCheckBoxChecked := FCheckBox.Checked;
  end;
  zBalloonControl.FShowMode := smDefault;
  if Assigned(zBalloonControl.OnClose) then
    zBalloonControl.OnClose(zBalloonControl);
end;

function Balloon_Key_HooK(code: Integer; wparam: WPARAM; lparam: LPARAM): LRESULT stdcall;
var Clickable: Boolean;
begin
 if (Code = 0) and ((lParam and $40000000) = 0) then begin
   Clickable := BalloonForm.lblClickHere.Visible;
   if Assigned(BalloonForm.OnClose) then
     BalloonForm.Close
   else BalloonForm.Release;
   Application.ProcessMessages;
   if WPARAM = VK_ESCAPE then
     Result := 1
   else begin
     if (zBalloonControl.FShowMode <> smHint) and
        Clickable and (WParam = VK_RETURN) and
        Assigned(zBalloonControl.OnClick) then begin
       Result := 1;
       zBalloonControl.OnClick(zBalloonControl);
     end
     else Result := 0;
   end;
 end
 else Result := CallNextHookEx(BalloonKeyboardHookHandle, Code, wParam, lParam);
end;

function Balloon_Mouse_HookProc(Code:integer; wParam: WPARAM; lParam: LPARAM): LRESULT stdcall;
var Clicked: Boolean;
begin
   Clicked := Assigned(zBalloonControl.OnClick) and
             (zBalloonControl.FShowMode = smDefault) and
              BalloonForm.lblClickHere.Visible and
             (wParam=WM_LBUTTONUP) and
             (PtInRect(BalloonForm.lblClickHere.BoundsRect,
                       BalloonForm.ScreenToClient(Mouse.CursorPos)));

   if (wParam = WM_LBUTTONUP) or (wParam = WM_LBUTTONDBLCLK) or
      (wParam = WM_RBUTTONUP) or (wParam = WM_RBUTTONDBLCLK) or
      (wParam = WM_MBUTTONUP) or (wParam = WM_MBUTTONDBLCLK) or
      (wParam = WM_RBUTTONDOWN) or (wParam = WM_MBUTTONDOWN) or
      (((wParam = WM_MOUSEWHEEL) or (wParam = WM_MOUSEMOVE)) and
      (not BalloonForm.tmrExit.Enabled) or
      (zBalloonControl.FShowMode in [smMessage, smError, smWorning]) or
       Clicked) then begin

     if (wParam <> WM_LBUTTONUP) and
//        (zBalloonControl.FShowMode in [smMessage, smError, smWorning]) and
        BalloonForm.tmrExit.Enabled {and
        //not PtInRect(BalloonForm.BoundsRect, BalloonForm.ScreenToClient(Mouse.CursorPos)) then begin
        not PtInRect(BalloonForm.BoundsRect, Mouse.CursorPos)} then begin
       Result := 0;
       Exit;
     end;

     if Assigned(BalloonForm.OnClose) then
         BalloonForm.Close
     else BalloonForm.Release;
     Application.ProcessMessages;
     if Clicked then begin
       Result := 0;
       zBalloonControl.OnClick(zBalloonControl);
     end
     else Result := 1;
   end
   else result:=CallNextHookEx(BalloonMouseHookHandle, Code, wParam, lParam);
end;

Procedure TBalloon.FormPaint(Sender: TObject);

  procedure HorzTwoWay(aTopColor, aEndColor: TColor);
  const mxcolors=100;
  type TGrArray=array[1..mxcolors] of tcolor;
  var
    Banda : TRect;
    q,I         : Integer;
    ga:tgrarray;

    procedure CalcColGradients(col1,col2:tcolor;n:integer);
    var i:integer;
         ar,ag,ab,br,bb,bg:integer;
         kr,kg,kb:real;
    begin
       ar:=GetRValue(col1);br:=GetRValue(col2);
       AG:=GetGValue(col1);bG:=GetGValue(col2);
       AB:=GetBValue(col1);bB:=GetBValue(col2);
       kr:=(br-ar) /n;
       kg:=(bg-ag) /n;
       kb:=(bb-ab) /n;
       for i:=0 to n-1 do
       ga[i+1]:=rgb(trunc(ar + i*kr),trunc(ag+i*kg),trunc(ab+i*kb));
    end;

  begin
   CalcColGradients(aEndColor, aTopColor, mxcolors);

   Banda.Left := 0;
   Banda.Right := Canvas.ClipRect.Right;
   q:= mxcolors div 2;
   with canvas do
   for I := 0 to q-1 do
   begin
       Banda.Top    := MulDiv (I, ClipRect.Bottom, mxcolors);
       Banda.Bottom := MulDiv (I + 1, ClipRect.Bottom, mxcolors);
       Brush.Color := ga[2*(i+1)];
       FillRect (Banda);
   end;
   with canvas do for I := q to mxcolors-1 do
   begin
       Banda.Top    := MulDiv (I    , ClipRect.Bottom, mxcolors);
       Banda.Bottom := MulDiv (I + 1, ClipRect.Bottom, mxcolors);
       Brush.Color := ga[mxcolors+1 - 2*(i-q+1)];
       FillRect (Banda);
   end;
  end;

Var
  TempRegion: HRGN;
  BM: TBitmap;
  x, y, v, h: Integer;
  aTopColor, aEndColor: TColor;
  aOptions: TCustomOptions;
Begin
  aTopColor := -1;
  aEndColor := clSilver;
  BM := nil;
  with zBalloonControl do
    case FShowMode of
      smDefault: aOptions := FOptions;
      smHint: aOptions := FHintOptions;
      smInput: aOptions := FInputOptions;
      smDialog: aOptions := FDlgOptions;
      else aOptions := FMsgOptions;
    end;

  if Assigned(aOptions.FWallPaper) then
    BM := aOptions.FWallPaper
  else begin
    aTopColor := aOptions.FTopColor;
    aEndColor := aOptions.FBottomColor;
  end;

  if Assigned(BM) then begin
    v := Height div BM.Height + 1;
    h := Width div BM.Width + 1;
    for x := 0 to h do
      for y := 0 to v do
        Canvas.Draw(x * BM.Width,
                    y * BM.Height,
                    BM);
  end
  else if aTopColor <> -1 then
    HorzTwoWay(aTopColor, aEndColor);
    //DrawGradient(Handle, aTopColor, aEndColor);

  with Canvas.Brush Do
   Begin
    Color := clBlack;
    Style := bsSolid;
   End;

  TempRegion := CreateRectRgn(0,0,1,1);
  GetWindowRgn(Handle, TempRegion);
  FrameRgn(Canvas.Handle, TempRegion, Canvas.Brush.handle, 1, 1);
  DeleteObject(TempRegion);
  if zBalloonControl.FShowMode in [smInput, smDialog] then OnPaint := nil;
End;

Procedure TBalloon.CreateParams(Var Params: TCreateParams);
const
  CS_DROPSHADOW = $00020000;
Begin
  Inherited CreateParams(Params);

  Params.Style     := (Params.Style and not WS_CAPTION) or WS_POPUP;
  Params.ExStyle   := Params.ExStyle or WS_EX_TOOLWINDOW or WS_EX_NOACTIVATE or WS_EX_TOPMOST;
  if (Win32MajorVersion = 5) and (Win32MinorVersion = 1) then
    Params.WindowClass.Style := Params.WindowClass.Style or CS_DROPSHADOW;
End;

Procedure TBalloon.OnChange(Sender: TObject);
Begin
//  if (csDestroying in Self.ComponentState) then Exit;
  SetWindowPos(Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOSIZE);
End;

Procedure TBalloon.WndProc(Var Message: TMessage);
Begin
//  if (csDestroying in ComponentState) then Exit;
  if (Message.Msg = WM_SIZE) and (Message.WParam = SIZE_MINIMIZED) Then
    Show;

  Inherited;
End;

procedure TBalloon.InitializeBalloon(X, Y: Integer; ArrowShow: Boolean);

  procedure MoveMouseOverControl(Sender: TObject);
  var point: TPoint;
  begin
    with TControl(Sender) do
    begin
      point.x:= Self.Left + left + (width div 5) * 4;
      point.y:= Self.Top + top + (height div 4) * 3;
      //point := Self.ClientToScreen(point);
      SetCursorPos(point.x, point.y);
    end;
  end;

var
  ArrowHeight, ArrowWidth, blnPosition, BtnTop, CharWidth, i: Integer;
  FormRegion, ArrowRegion: HRGN;
  Arrow: Array [0..2] Of TPoint;
  InTop, InLeft: Boolean;
  s, v: string;
  bImage: TImage;
  bRadio: TzzHtmlLabel;

  procedure SetFinishedHeight;
  begin
    if lblClickHere.Visible then begin
      lblClickHere.Width := lblText.Width;
      lblClickHere.Caption := zBalloonControl.FClickHereText;
      lblClickHere.AutoSize := True;
      lblClickHere.Left := lblText.Left;
      lblClickHere.Update;
      lblClickHere.Top := Height;
      Height := Height + lblClickHere.Height + 5;
      If blnPosition < 2 Then
        lblClickHere.Top := Height - lblClickHere.Height - 5 - ArrowHeight;
    end;
  end;

  procedure Icon2Bitmap(const Icon: TIcon; Bitmap: TBitmap);
  var TempBitmap: TBitmap;
      IconInfo: TIconInfo;
  begin
    // Retrieve icon data
    if (not GetIconInfo(Icon.Handle, IconInfo)) then Exit;
    TempBitmap := TBitmap.Create;
    try
      // Borrow the icon bitmap handle and assign it to a TBitmap
      TempBitmap.Handle := IconInfo.hbmColor;
      try
        // Copy bitmap
        Bitmap.Assign(TempBitmap);
        // Force copy to clone handle
        Bitmap.Handle;
      finally
        // Give the bitmap handle back to the icon
        TempBitmap.ReleaseHandle;
      end;
    finally
      TempBitmap.Free;
    end;
  end;

begin
  ArrowHeight := 20;
  ArrowWidth  := 20;
                          
  if zBalloonControl.FShowMode in [smMessage, smError, smWorning, smDialog, smInput] then begin
    case zBalloonControl.FShowMode of
      smInput: if zBalloonControl.FInputType = itPassword then
                 IconBitmap.Picture.Bitmap.LoadFromResourceName(HInstance,'BITMAP_LOCK')
               else IconBitmap.Picture.Bitmap.LoadFromResourceName(HInstance,'BITMAP_INPUT');
      smDialog: if zBalloonControl.FButtonType = btOK then
                  IconBitmap.Picture.Bitmap.LoadFromResourceName(HInstance,'BITMAP_INFO')
                else IconBitmap.Picture.Bitmap.LoadFromResourceName(HInstance,'BITMAP_HELP');
      smMessage: IconBitmap.Picture.Bitmap.LoadFromResourceName(HInstance,'BITMAP_INFO');
      smError: IconBitmap.Picture.Bitmap.LoadFromResourceName(HInstance,'BITMAP_ERROR');
      smWorning: IconBitmap.Picture.Bitmap.LoadFromResourceName(HInstance,'BITMAP_WORNING');
    end;
    IconBitmap.Top := lblTitle.Top;
  end
  else case zBalloonControl.FIconMode of
    imNone: IconBitmap.Visible := False;
    imGlyph: begin
               if Assigned(zBalloonControl.FGlyph) then
                 IconBitmap.Picture.Assign(zBalloonControl.FGlyph)
               else IconBitmap.Visible := False;
             end;
    imImageList: begin
                   if Assigned(zBalloonControl.FImageList) and
                      (zBalloonControl.FImageIndex <> -1) and
                      (zBalloonControl.FImageIndex in
                       [0..zBalloonControl.FImageList.Count - 1]) then
                     zBalloonControl.FImageList.GetBitmap(zBalloonControl.FImageIndex,
                            IconBitmap.Picture.Bitmap)
                   else IconBitmap.Visible := False;
                 end;
    imApplIcon: Icon2Bitmap(Application.Icon, IconBitmap.Picture.Bitmap);
  end;

  if IconBitmap.Visible then
    lblTitle.Left := IconBitmap.Left * 2 + IconBitmap.Width
  else lblTitle.Left := IconBitmap.Left;
  lblTitle.Width := width - lblTitle.Left - 10;
  if zBalloonControl.FTitle <> '' then
    lblTitle.Caption := zBalloonControl.FTitle
  else lblTitle.Caption := Application.Title;
  lblTitle.Width := lblTitle.MaxLineLength;
  lblTitle.Paint;

  if zBalloonControl.FShowMode in [smMessage, smError, smWorning, smInput, smDialog] then begin
    lblText.Width := zBalloonControl.MaxWidth - lblText.Left;
    lblText.Left := lblTitle.Left;
    lblText.Top := lblTitle.Top + lblTitle.Height + 8;
  end
  else begin
    if lblTitle.Height > IconBitmap.Height then
      lblText.Top := lblTitle.Top + lblTitle.Height + 8
    else lblText.Top := IconBitmap.Top + IconBitmap.Height + 4;
    if Length(zBalloonControl.FText.Text) > (Screen.Width div 3) then begin
      lblText.Width := Screen.Width div 3;
      lblTitle.Width := lblText.Width - lblTitle.Left - 10;
    end
    else lblText.Width := zBalloonControl.MaxWidth - lblText.Left * 2;
  end;
  lblText.Caption := Trim(zBalloonControl.FText.Text);
  lblText.Width := lblText.MaxLineLength;
  lblText.Paint;

  if (lblTitle.Left + lblTitle.Width) > (lblText.Left + lblText.Width) then
    Width := lblTitle.Left + lblTitle.Width + 10
  else Width := lblText.Left + lblText.MaxLineLength + 10;

  FormRegion := 0;

  InTop := Y > (Screen.Height div 2);
  InLeft := X > (Screen.Width div 2);

  if InTop and InLeft then
    blnPosition := 1
  else if InTop and not InLeft then
    blnPosition := 0
  else if not InTop and InLeft then
    blnPosition := 2
  else blnPosition := 3;

  if not ArrowShow then begin
    Height := lblText.Top + lblText.Height + 8;
    BtnTop := lblText.Top + lblText.Height + 8;
    if zBalloonControl.FShowMode in [smInput, smDialog] then begin
      if zBalloonControl.FShowMode = smInput then begin
        if zBalloonControl.FInputType in [itCombo, itComboList] then begin
          FComboBox := TComboBox.Create(Self);
          with FComboBox do begin
            Parent := Self;
            Font.Assign(zBalloonControl.FInputOptions.FEdit.FFont);
            Color := zBalloonControl.FInputOptions.FEdit.FColor;
            Top := lblText.Top + lblText.Height + 4;
            Left := lblText.Left;
            Width := Round(Width * 1.5);
            MaxLength := Round(zBalloonControl.FInputOptions.FMax);
            if zBalloonControl.FInputType = itComboList then
              Style := csDropDownList;
            if zBalloonControl.FInputValue <> '' then
              Text := zBalloonControl.FInputValue
            else if zBalloonControl.FInputOptions.fRadioIndex <> -1 then
              ItemIndex := zBalloonControl.FInputOptions.fRadioIndex;
            SelStart := Length(Text);
          end;
          if FComboBox.Left + FComboBox.Width + 8 > Width then
            Width := FComboBox.Left + FComboBox.Width + 8;
          BtnTop := FComboBox.Top + FComboBox.Height + 16;
        end
        else if zBalloonControl.FInputType in [itCheckBox, itRadioButton] then begin
          for i := 0 to Pred(zBalloonControl.InputOptions.fBoxList.Count) do begin
            s := zBalloonControl.InputOptions.fBoxList.Names[i];
            {$IFDEF VER150}  // Michal Mutl
            if s <> '' then v := Lowercase(zBalloonControl.InputOptions.fBoxList.ValueFromIndex[i])
            {$ELSE}
            if s <> '' then v := Lowercase(zBalloonControl.InputOptions.fBoxList.Values[zBalloonControl.InputOptions.fBoxList.Names[i]])
            {$ENDIF}
            else begin
              s := zBalloonControl.InputOptions.fBoxList[i];
              v := '';
            end;
            bImage := TImage.Create(Self);
            with bImage do begin
              Parent      := Self;
              AutoSize := True;
              Top         := BtnTop + 1;
              Left        := lblText.Left;
              OnClick := OnCheckBoxClick;
            end;
            bRadio := TzzHtmlLabel.Create(Self);
            with bRadio do begin
              Parent := Self;
              Name := 'zCheck' + IntToStr(i);
              Transparent := True;
              Left := lblText.Left + 20;
              Top := BtnTop;
              Width := Self.Width - Left - 8;
              Caption := s;
              FBox := bImage;
              Tag := i;
              FChecked := (v = 'true') or (v = 'yes') or (v = '1');
              if FChecked then begin
                if zBalloonControl.FInputType = itCheckBox then
                  FBox.Picture.Bitmap.LoadFromResourceName(HInstance, 'Select')
                else if not Assigned(FRadioChecked) then begin
                  FBox.Picture.Bitmap.LoadFromResourceName(HInstance, 'Checked');
                  FRadioChecked := bRadio;
                end
                else FBox.Picture.Bitmap.LoadFromResourceName(HInstance, 'Unselect');
              end
              else FBox.Picture.Bitmap.LoadFromResourceName(HInstance, 'Unselect');
              OnClick := OnCheckBoxClick;
              BtnTop := BtnTop + Height + 1;
            end;
          end;
          BtnTop := BtnTop + 12;
        end
        else begin
          FEdit := TEdit.Create(Self);
          with FEdit do begin
            Parent := Self;
            Font.Assign(zBalloonControl.FInputOptions.FEdit.FFont);
            Color := zBalloonControl.FInputOptions.FEdit.FColor;
            Top := lblText.Top + lblText.Height + 4;
            Left := lblText.Left;
            Width := Round(Width * 1.5);
            if (zBalloonControl.FInputType in [itString, itPassword]) and
               (zBalloonControl.FInputOptions.FMax > 0) then
              MaxLength := Round(zBalloonControl.FInputOptions.FMax);
            if zBalloonControl.FInputType = itPassword then
              PasswordChar := '*';
            Text := zBalloonControl.FInputValue;
            SelStart := Length(Text);
          end;
          if FEdit.Left + FEdit.Width + 8 > Width then
            Width := FEdit.Left + FEdit.Width + 8
          else FEdit.Width := Width - FEdit.Left - 8;

          BtnTop := FEdit.Top + FEdit.Height + 16;
        end;
      end
      else BtnTop := lblText.Top + lblText.Height + 16;

      CharWidth := Canvas.TextWidth('u') * 2;
      if zBalloonControl.FShowMode = smInput then
        zBalloonControl.FButtonOptions := zBalloonControl.FInputOptions.FButtonOptions
      else zBalloonControl.FButtonOptions := zBalloonControl.FDlgOptions.FButtonOptions;

      // Yes/Ok button
      Btn1 := TzzButton.Create(Self);
      with Btn1 do begin
        Parent := Self;
        Top := BtnTop;
        if zBalloonControl.FButtonType in [btOC, btOK] then
          Caption := zBalloonControl.FButtonCaptions.FOk
        else Caption := zBalloonControl.FButtonCaptions.FYes;
        if Canvas.TextWidth(Caption) + CharWidth > Width then
          Width := Canvas.TextWidth(Caption) + CharWidth;
        ModalResult := mrYes;
        OnClick := OnYesClick;
      end;

      // No Button
      if zBalloonControl.FButtonType = btYNC then begin
        Btn2 := TzzButton.Create(Self);
        with Btn2 do begin
          Parent := Self;
          Top := BtnTop;
          Caption := zBalloonControl.FButtonCaptions.FNo;
          if Canvas.TextWidth(Caption) + CharWidth > Width then
            Width := Canvas.TextWidth(Caption) + CharWidth;
          ModalResult := mrNo;
        end;
      end;

      if zBalloonControl.FButtonType <> btOK then begin
        // Cancel Button
        Btn3 := TzzButton.Create(Self);
        with Btn3 do begin
          Parent := Self;
          Top := BtnTop;
          if zBalloonControl.FButtonType = btYN then
            Caption := zBalloonControl.FButtonCaptions.FNo
          else Caption := zBalloonControl.FButtonCaptions.FCancel;
          if Canvas.TextWidth(Caption) + CharWidth > Width then
            Width := Canvas.TextWidth(Caption) + CharWidth;
          Cancel := True;
          ModalResult := mrCancel;
        end;

        CharWidth := Btn1.Width + Btn3.Width + 1;
        if zBalloonControl.FButtonType = btYNC then
          Inc(CharWidth, Btn2.Width + 1);
        if CharWidth + lblText.Left + 8 > Width then
          Width := CharWidth + lblText.Left + 8;

        Btn3.Left := Width - 8 - Btn3.Width;
        if zBalloonControl.FButtonType = btYNC then begin
          Btn2.Left := Btn3.Left - Btn2.Width - 1;
          Btn1.Left := Btn2.Left - Btn1.Width - 1;
        end
        else Btn1.Left := Btn3.Left - Btn1.Width - 1;
      end
      else Btn1.Left := Width - 8 - Btn1.Width;

      Height := Btn1.Top + Btn1.Height + 8;

      if zBalloonControl.FShowMode = smInput then begin
        if not (zBalloonControl.FInputType in [itCheckBox, itRadioButton, itCombo, itComboList]) then begin
          FEdit.OnChange := OnEditChange;
          FEdit.OnKeyPress := OnEditKeyPress;
          OnEditChange(nil);
        end;
        if zBalloonControl.FInputOptions.FCheckBoxCaption <> '' then begin
          FCheckBox := TzBalloonCheckBox.Create(Self);
          with FCheckBox do begin
            Parent := Self;
            Caption := zBalloonControl.FInputOptions.FCheckBoxCaption;
            Checked := zBalloonControl.FInputOptions.FCheckBoxChecked;
          end;
        end;
      end
      else if zBalloonControl.FDlgOptions.FCheckBoxCaption <> '' then begin
        FCheckBox := TzBalloonCheckBox.Create(Self);
        with FCheckBox do begin
          Parent := Self;
          Caption := zBalloonControl.FDlgOptions.FCheckBoxCaption;
          Checked := zBalloonControl.FDlgOptions.FCheckBoxChecked;
        end;
      end;
    end;
    if Assigned(FCheckBox) then begin
      with FCheckBox do begin
        Top := Btn1.Top + Btn1.Height + 8;
        Left := lblText.Left;
        lblClickHere.Width := lblText.Width - 20;
        lblClickHere.Caption := Caption;
        lblClickHere.AutoSize := True;
        lblClickHere.Update;
        Height := lblClickHere.Height;
        Width := Self.Width - Left - 8;
        lblClickHere.Caption := '';
      end;
      Height := FCheckBox.Top + FCheckBox.Height + 8;
    end;
    SetFinishedHeight;
    Left := X - Width div 2;
    Top  := Y - Height div 2;

    {case blnPosition of
      0: begin
           Left := X;
           Top  := Y - Height;
         end;
      1: Begin
           Left := X - Width;
           Top  := Y - Height;
         end;
      2: Begin
           Left := X - (Width - 20);
           Top  := Y;
         end;
      3: Begin
           Left := X - Width div 2;
           Top  := Y - Height div 2;
         end;
    end;}
    FormRegion := CreateRoundRectRgn(0, 0, Width, Height, 7, 7);
  end
  else begin
    If blnPosition > 1 Then begin
      lblTitle.Top := lblTitle.Top + ArrowHeight;
      IconBitmap.Top := IconBitmap.Top + ArrowHeight;
      if lblTitle.Height > IconBitmap.Height then
        lblText.Top := lblTitle.Top + lblTitle.Height + 8
      else lblText.Top := IconBitmap.Top + IconBitmap.Height + 4;
      Height := lblText.Top + lblText.Height + 8;
    end
    else Height := lblText.Top + lblText.Height + 8 + ArrowHeight;
    SetFinishedHeight;

    case blnPosition of
      0: Begin
           Left := X - 20;
           Top  := Y - Height;
           FormRegion := CreateRoundRectRgn(0, 0, Width, Height - (ArrowHeight - 2), 7, 7);

           Arrow[0] := Point(20, Height - ArrowHeight);
           Arrow[1] := Point(20, Height);
           Arrow[2] := Point(20 + ArrowWidth, Height - ArrowHeight);
         End;
      1: Begin
           Left := X - (Width - 20);
           Top  := Y - Height;
           FormRegion := CreateRoundRectRgn(0, 0, Width, Height - (ArrowHeight - 2), 7, 7);

           Arrow[0] := Point(Width - ArrowWidth - 20, Height - ArrowHeight);
           Arrow[1] := Point(Width - 20, Height);
           Arrow[2] := Point(Width - 20, Height - ArrowHeight);
         End;
      2: Begin
           Left := X - (Width - 20);
           Top  := Y - 2;
           FormRegion := CreateRoundRectRgn(0, ArrowHeight + 2, Width, Height, 7, 7);

           Arrow[0] := Point(Width - 20, 2);
           Arrow[1] := Point(Width - 20, ArrowHeight + 2);
           Arrow[2] := Point(Width - 20 - ArrowWidth, ArrowHeight + 2);
         End;
      3: Begin
           Left := X - 20;
           Top  := Y - 2;
           FormRegion := CreateRoundRectRgn(0, ArrowHeight + 2, Width, Height, 7, 7);

           Arrow[0] := Point(20, 2);
           Arrow[1] := Point(20, ArrowHeight + 2);
           Arrow[2] := Point(20 + ArrowWidth, ArrowHeight + 2);
         End;
    end;
  end;

  If ArrowShow Then begin
    ArrowRegion := CreatePolygonRgn(Arrow, 3, WINDING);
    CombineRgn(FormRegion, FormRegion, ArrowRegion, RGN_OR);
    DeleteObject(ArrowRegion);
  end;
  SetWindowRgn(Handle, FormRegion, True);

  Visible := False;
  if zBalloonControl.FShowMode in [smInput, smDialog, smMessage, smError, smWorning] then begin
    OnClose := FormClose;
    if Assigned(Btn1) then
      MoveMouseOverControl(Btn1);
    if zBalloonControl.FShowMode in [smMessage, smError, smWorning] then begin
      tmrExit.Enabled := True;
      ShowModal;
    end
    else zBalloonControl.FInputResult := ShowModal;
  end
  else begin
    ShowWindow(Handle, SW_SHOWNOACTIVATE);
    Visible := True;
    tmrExit.Enabled := True;
  end;
  OldOnActiveControlChange := Screen.OnActiveControlChange;
  Screen.OnActiveControlChange := OnActiveControlChange;
  OldOnActiveFormChange := Screen.OnActiveFormChange;
  Screen.OnActiveFormChange := OnActiveFormChange;
  Refresh;
end;

procedure TBalloon.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action := caFree;
end;

procedure TBalloon.OnYesClick(sender : TObject);
var i: Integer;
    s: string;
begin
  if zBalloonControl.FInputType in [itCheckBox, itRadioButton] then begin
    if zBalloonControl.FInputType = itCheckBox then begin
      for i := 0 to ControlCount - 1 do begin
        if Controls[i] is TzzHtmlLabel then
          with Controls[i] as TzzHtmlLabel do
            if Pos('zCheck', Name) > 0 then begin
              s := zBalloonControl.InputOptions.fBoxList.Names[Tag];
              if s = '' then
                s := zBalloonControl.InputOptions.fBoxList[Tag];
              if Checked then
                zBalloonControl.InputOptions.fBoxList[Tag] := s + '=1'
              else zBalloonControl.InputOptions.fBoxList[Tag] := s + '=0';
            end;
      end;
    end;
  end
  else if zBalloonControl.FInputType in [itCombo, itComboList] then
    zBalloonControl.FInputValue := FComboBox.Text
  else zBalloonControl.FInputValue := FEdit.Text;
end;

procedure TBalloon.OnCheckBoxClick(Sender : TObject);
var zzLabel: TzzHtmlLabel;
    i: Integer;
begin
  zzLabel := nil;
  if Sender is TzzHtmlLabel then
    zzLabel := Sender as TzzHtmlLabel
  else begin
    for i := 0 to ControlCount - 1 do
      if Controls[i] is TzzHtmlLabel then
        with Controls[i] as TzzHtmlLabel do
          if (Pos('zCheck', Name) > 0) and (FBox = Sender) then
            zzLabel := (Controls[i] as TzzHtmlLabel);
    if not Assigned(zzLabel) then Exit;
  end;

  if zBalloonControl.FInputType = itRadioButton then begin
    if Assigned(FRadioChecked) then begin
      if FRadioChecked = zzLabel then Exit;
      FRadioChecked.Checked := False;
    end;
    FRadioChecked := zzLabel;
    FRadioChecked.Checked := True;
    zBalloonControl.InputOptions.fRadioIndex := FRadioChecked.Tag;
  end
  else with zzLabel do Checked := not Checked;
end;

procedure TBalloon.OnTimer(Sender: TObject);
begin
  if tmrExit.Interval = 100 then begin
    if (zBalloonControl.FShowMode = smHint) then begin
      tmrExit.Interval := Application.HintHidePause;
    end
    else begin
      tmrExit.Interval := (Length(lblTitle.Caption) + Length(lblText.Caption)) * 50;
      if tmrExit.Interval < 2000 then
        tmrExit.Interval := 2000;
{      OnClick          := OnMouseClick;
      iconBitmap.OnClick := OnMouseClick;
      pnlAlign.OnClick := OnMouseClick;
      lblTitle.OnClick := OnMouseClick;
      lblText.OnClick  := OnMouseClick;}
    end;
    if not (CSDesigning in ComponentState) then begin
      BalloonKeyboardHookHandle := SetWindowsHookEx(WH_KEYBOARD, Balloon_Key_HooK, hInstance, GetCurrentThreadID);
      BalloonMouseHookHandle := SetWindowsHookEx(WH_MOUSE, Balloon_Mouse_HookProc, HInstance, GetCurrentThreadID);
    end;
  end
  else begin
    tmrExit.Enabled := False;
    if (zBalloonControl.FShowMode = smHint) then begin
      BalloonForm.Release;
      Application.ProcessMessages;
    end
{    else if not (CSDesigning in ComponentState) then begin
//      BalloonKeyboardHookHandle := SetWindowsHookEx(WH_KEYBOARD, Balloon_Key_HooK, hInstance, GetCurrentThreadID);
      BalloonMouseHookHandle := SetWindowsHookEx(WH_MOUSE, Balloon_Mouse_HookProc, HInstance, GetCurrentThreadID);
    end;}
  end;
end;

procedure TBalloon.OnActiveControlChange(sender : TObject);
begin
  Close;
  if not (csDestroying in ComponentState) and Assigned(OldOnActiveControlChange) then
    OldOnActiveControlChange(sender);
end;

procedure TBalloon.OnActiveFormChange(sender : TObject);
begin
  Close;
  if not (csDestroying in ComponentState) and Assigned(OldOnActiveFormChange) then
    OnActiveFormChange(sender);
end;

procedure TBalloon.WMNCHitTest(var M: TWMNCHitTest);
begin
  inherited;
  if (zBalloonControl.FShowMode in [smInput, smDialog]) and
     not (zBalloonControl.FInputType in [itCheckBox, itRadioButton]) then begin
    if M.Result = htClient then M.Result := htCaption;
  end;
end;

Procedure TBalloon.OnEditChange(Sender: TObject);
var namePart, serverPart: string;
    i: Integer;
begin
  Btn1.Enabled := True;
  case zBalloonControl.FInputType of
    itPassword,
    itString: begin
                if zBalloonControl.FInputOptions.FMin > 0 then
                  Btn1.Enabled := Length(FEdit.Text) >= Round(zBalloonControl.FInputOptions.FMin);
                if Btn1.Enabled and (zBalloonControl.FInputOptions.FMax > 0) and
                   (zBalloonControl.FInputOptions.FMax > zBalloonControl.FInputOptions.FMin) then
                  Btn1.Enabled := Length(FEdit.Text) <= Round(zBalloonControl.FInputOptions.FMax);
              end;
    itInteger: begin
                 Btn1.Enabled := (FEdit.Text <> '') and (FEdit.Text <> '-');
                 if Btn1.Enabled and (zBalloonControl.FInputOptions.FMin <> 0) then
                   Btn1.Enabled := StrToInt(FEdit.Text) >= Round(zBalloonControl.FInputOptions.FMin);
                 if Btn1.Enabled and (zBalloonControl.FInputOptions.FMax <> 0) and
                   (zBalloonControl.FInputOptions.FMax > zBalloonControl.FInputOptions.FMin) then
                  Btn1.Enabled := StrToInt(FEdit.Text) <= Round(zBalloonControl.FInputOptions.FMax);
               end;
    itMail: begin
              i := Pos('@', FEdit.Text);
              if i = 0 then begin
                Btn1.Enabled := False;
                Exit;
              end;
              namePart:= Copy(FEdit.Text, 1, i - 1);
              serverPart:= Copy(FEdit.Text, i + 1, MaxInt);
              i:= Pos('.', serverPart);
              if (Length(namePart) = 0) or ((Length(serverPart) < 5)) then begin
                Btn1.Enabled := False;
                Exit;
              end;
              Btn1.Enabled := (i = 0) or (i > (Length(serverPart) - 2));
            end;
     itURL: begin
              namePart := LowerCase(FEdit.Text);
              i := LastDelimiter('.', namePart);
              if ((Pos('http://', namePart) <> 1) and (Pos('www.', namePart) <> 1)) or
                 (i = 0) or
                 ((Pos('www.', namePart) > 0) and (Pos('.', namePart) = i)) or
                 (i >= Length(namePart) - 1) then
                Btn1.Enabled := False
              else Btn1.Enabled := True;
            end;

  end;
  Btn1.Invalidate;
end;

{procedure TBalloon.OnMouseClick(Sender: TObject);
begin
  Close;
end;}

procedure TBalloon.OnEditKeyPress(Sender: TObject; var Key: Char);
var p: Integer;
begin
  if Key = Chr(13) then begin
    Key := Chr(0);
    if Btn1.Enabled then Btn1.Click;
    Exit;
  end
  else begin
    if Key < ' ' then Exit;
    case zBalloonControl.FInputType of
      itInteger,
      itDouble: begin
                  if not (Key in [DecimalSeparator, '-', '0'..'9']) then begin  // #8 is Backspace
                    Key := Chr(0);
                    Exit;
                  end;

                  p := Pos(DecimalSeparator, FEdit.Text);

                  if (Key = '-') then begin
                    if (FEdit.SelLength <> Length(FEdit.Text)) or
                       (FEdit.SelStart <> 0) or
                       ((Length(FEdit.Text) > 0) and (FEdit.Text[1] = '-')) or
                       (zBalloonControl.FInputOptions.FMin > 0) then
                      Key := Chr(0);
                  end
                  else if (Key = DecimalSeparator) then begin
                    if (zBalloonControl.FInputType = itInteger) then
                      Key := Chr(0)
                    else if (FEdit.SelLength = Length(FEdit.Text)) or
                       (FEdit.SelStart = 0) or
                       (zBalloonControl.FInputOptions.FDecimalCount = 0) or
                       (p > 0) or
                       (Length(FEdit.Text) - FEdit.SelStart > zBalloonControl.FInputOptions.FDecimalCount) then
                      Key := Chr(0);
                  end
                  else if (zBalloonControl.FInputType = itDouble) then begin
                    if (p > 0) and
                       (Length(FEdit.Text) - p >=
                        zBalloonControl.FInputOptions.FDecimalCount) and
                       (FEdit.SelStart > p) then
                      Key := Chr(0);
                  end;
                end;
      itFileName: if Pos(Key, cIllegalChars) > 0 then
                    Key := Chr(0);
      itMail: if not (Key in ['a'..'z', 'A'..'Z', '0'..'9', '_', '-', '.', '@']) then
                Key := Chr(0);
      itURL: if not (Key in ['a'..'z', 'A'..'Z', '0'..'9', '_', '-', '.', '/']) then
               Key := Chr(0);
    end;
  end;

  if (Key <> Chr(0)) and Assigned(zBalloonControl.FKeyPressEvent) then
    zBalloonControl.FKeyPressEvent(FEdit, Key);
end;

///////////////////////////////////////// zzHtmlLabel

constructor TzzHtmlLabel.Create(AOwner: TComponent);
begin
  inherited;
  AutoSize:=False;
  FDrawing := False;
  FBox := nil;
end;

procedure TzzHtmlLabel.Paint;
begin
  DrawStringEx;
end;

procedure TzzHtmlLabel.SetCaption(Value: string);
begin
  if Value <> FCaption then begin
    MaxLineLength := 0;
    NeedHeight := 0;
    FCaption := Value;
    DrawStringEx;
  end;
end;

Procedure TzzHtmlLabel.DrawStringEx;
var oldColor: TColor;
    oldStyle: TFontStyles;
    i, k: Integer;
    s, sCmd,sAux: String;
    rctNow: TRect;
    Finished: Boolean;

  Procedure SetBold(b: Boolean);
  begin
    If B then Canvas.Font.Style:=Canvas.Font.Style+[fsBold]
      else Canvas.Font.Style:=Canvas.Font.Style-[fsBold];
  end;

  Procedure SetItalic(b: Boolean);
  begin
    If B then Canvas.Font.Style:=Canvas.Font.Style+[fsItalic]
      else Canvas.Font.Style:=Canvas.Font.Style-[fsItalic];
  end;

  Procedure SetUnderline(b: Boolean);
  begin
    If B then Canvas.Font.Style:=Canvas.Font.Style+[fsUnderline]
      else Canvas.Font.Style:=Canvas.Font.Style-[fsUnderline];
  end;

  Procedure DrawStr(S: String);
  var iW, ii, l: Integer;
      sAux1, sA: String;
      IsSeparator: Boolean;
  begin
    while s <> '' do begin
      iW := Canvas.TextWidth(s);
      If (rctNow.Bottom > Height) then
        Height := rctNow.Bottom;

      If iW <= rctNow.Right-rctNow.Left then begin
        Canvas.TextOut(rctNow.Left, rctNow.Top, S);
        rctNow.Left:=rctNow.Left+iW;
        S:='';
        if rctNow.Left > MaxLineLength then MaxLineLength := rctNow.Left;
      end
      else begin
        sAux1:='';
        l := Length(s);
        IsSeparator := False;
        for ii := Length(s) downto 1 do begin
          if Pos(s[ii], ' -\|+/') = 0 then
            sAux1 := s[ii] + sAux1
          else begin
            IsSeparator := True;
            sA := Copy(s, 1, l - Length(sAux1) - 1);
            iW := Canvas.TextWidth(sA);
            if iW <= rctNow.Right - rctNow.Left then Break;
            sAux1 := s[ii] + sAux1;
          end;
        end;

        if (sAux1 <> s) and (Length(sAux1) >= 2) then begin  // not all from new line
          iW := Canvas.TextWidth(sA);
          Canvas.TextOut(rctNow.Left, rctNow.Top, sA);
          if rctNow.Left + iW > MaxLineLength then MaxLineLength := rctNow.Left + iW;
          rctNow.Left:=ClientRect.Left;
          rctNow.Right:=ClientRect.Right;
          rctNow.Top:=rctNow.Bottom;
          rctNow.Bottom:=rctNow.Top + Canvas.TextHeight('Wg');
          Delete(s, 1, Length(sA));
          while (Length(s) > 0) and (s[1] = ' ') do
            Delete(s, 1, 1);
        end
        else if not IsSeparator and   // if not clipped
             (rctNow.Left = ClientRect.Left) then begin
          Canvas.TextRect(rctNow, rctNow.Left, rctNow.Top, s);
          iW := ClientRect.Right - ClientRect.Left;
          if iW > MaxLineLength then MaxLineLength := iW;
          s := '';
        end
        else begin
          if (sAux1 = s) then begin
            rctNow.Left:= ClientRect.Left;
            rctNow.Right:=ClientRect.Right;
            rctNow.Top:=rctNow.Bottom;
            rctNow.Bottom:=rctNow.Top + Canvas.TextHeight('Wg');
          end;
          Canvas.TextRect(rctNow, rctNow.Left, rctNow.Top, s);
          if (sAux1 <> s) and not Finished then begin
            rctNow.Left:=ClientRect.Left;
            rctNow.Right:=ClientRect.Right;
            rctNow.Top:=rctNow.Bottom;
            rctNow.Bottom:=rctNow.Top + Canvas.TextHeight('Wg');
          end
          else if (sAux1 = s) then begin
            rctNow.Left:= rctNow.Left + Canvas.TextWidth(s);
            rctNow.Right:=ClientRect.Right;
          end;
          s := '';
        end;
      end;
    end;
  end;

begin
  if FDrawing then Exit;
  FDrawing := True;
  try
    Canvas.Font.Assign(Font);
    Canvas.Brush.Color := Color;

    If Transparent then Canvas.Brush.Style:=bsClear
      else Canvas.Brush.Style:=bsSolid;

    Canvas.FillRect(ClientRect);

    i:=1;
    oldColor:=Canvas.Font.Color;
    oldStyle:=Canvas.Font.Style;

    s := TrimRight(Caption);
    s := StringReplace(s, #13#10, '<br>', [rfReplaceAll]);
    s := StringReplace(s, #13, '<br>', [rfReplaceAll]);
    s := StringReplace(s, #10, '<br>', [rfReplaceAll]);
{    if Pos('<BR>', UpperCase(s)) = 0 then begin
    end
    else begin
      if Pos(#13, s) > 0 then
        s := StringReplace(s, #13, '', [rfReplaceAll]);
      if Pos(#10, s) > 0 then
        s := StringReplace(s, #10, '', [rfReplaceAll]);
    end;}

    NeedHeight := Canvas.TextHeight('Wg');
    rctNow := Rect(ClientRect.Left, ClientRect.Top, ClientRect.Right, ClientRect.Top + Canvas.TextHeight('Wg'));
    while i <= Length(S) do begin
      if S[i]='<' then begin
        sCmd:='';
        Inc(i);
        while (i<=Length(S)) and (S[i]<>'>') do begin
          sCmd:=sCmd+UpCase(S[i]);
          Inc(i);
        end;
        Inc(i);
        if sCmd='<' then DrawStr('<')
          else begin
            If sCmd='BR' then begin
              rctNow.Left:=ClientRect.Left;
              if rctNow.Left > MaxLineLength then MaxLineLength := rctNow.Left;
              rctNow.Right:=ClientRect.Right;
              rctNow.Top:=rctNow.Bottom;
              rctNow.Bottom:=rctNow.Top+Canvas.TextHeight('Wg');
            end
            else If sCmd='LB' then begin
              if rctNow.Left <> ClientRect.Left then begin
                rctNow.Top:=rctNow.Bottom;
                rctNow.Bottom := rctNow.Top + Canvas.TextHeight('Wg');
              end;
              rctNow.Left:=ClientRect.Left;
              if rctNow.Left > MaxLineLength then MaxLineLength := rctNow.Left;
              rctNow.Right:=ClientRect.Right;
              Canvas.Pen.Color := Canvas.Font.Color;
              k := (rctNow.Bottom - rctNow.Top) div 2;
              Canvas.MoveTo(rctNow.Left, rctNow.Top + k);
              Canvas.LineTo(rctNow.Right, rctNow.Top + k);
              rctNow.Top:=rctNow.Bottom;
              rctNow.Bottom:=rctNow.Top+Canvas.TextHeight('Wg');
            end
            else If sCmd='B' then SetBold(True)
            else If sCmd='/B' then SetBold(False)
            else If sCmd='I' then SetItalic(True)
            else If sCmd='/I' then SetItalic(False)
            else If sCmd='U' then SetUnderline(True)
            else If sCmd='/U' then SetUnderline(False)
            else If sCmd='RU' then Canvas.Font.Charset:=RUSSIAN_CHARSET
            else If sCmd='EU' then Canvas.Font.Charset:=ANSI_CHARSET
            else If Pos('/C', sCmd) = 1 then Canvas.Font.Color:= Font.Color
            else If (Pos('C:',sCmd)=1) or (Pos('C=',sCmd)=1) then begin
              Delete(sCmd,1,2);
              try
                if Copy(sCmd,1,1)='$' then Canvas.Font.Color:=StringToColor(sCmd)
                else if Copy(sCmd,1,1)='#' then   // Html color
                  Canvas.Font.Color:=StringToColor('$'+copy(sCmd,6,2)+copy(sCmd,4,2)+copy(sCmd,2,2))
                else if Copy(sCmd, 1, 2) = 'CL' then
                  Canvas.Font.Color:=StringToColor(sCmd)
                else Canvas.Font.Color:=StringToColor('cl'+sCmd)
              except
              end;
            end;
          end;
      end
      else begin
        sAux:='';
        While (I<=Length(S)) and (S[I]<>'<') do begin
          sAux:=sAux+S[I];
          Inc(I);
        end;
        Finished := i > Length(S);
        DrawStr(sAux);
      end;
    end;
//    if (rctNow.Bottom = Height + Canvas.TextHeight('Wg') * 2) then
//      Height := Height + Canvas.TextHeight('Wg');
    if (rctNow.Bottom > Height) then
      Height := rctNow.Bottom;
    Canvas.Font.Color:=oldColor;
    Canvas.Font.Style:=oldStyle;
  finally
    NeedHeight := Height + 1;
    FDrawing := False;
  end;
end;

procedure TzzHtmlLabel.SetChecked(Value: Boolean);
begin
  if FChecked = Value then Exit;
  FChecked := Value;
  if FChecked then begin
    if zBalloonControl.FInputType = itCheckBox then
      FBox.Picture.Bitmap.LoadFromResourceName(HInstance, 'Select')
    else begin
      FBox.Picture.Bitmap.LoadFromResourceName(HInstance, 'Checked');
    end;
  end
  else FBox.Picture.Bitmap.LoadFromResourceName(HInstance, 'Unselect');
end;

////////////////////////////////  Custom Options

constructor TCustomOptions.Create(AOwner: TzBalloon);
begin
  inherited Create;
  FFont := TFont.Create;
  FTitleFont := TFont.Create;
  with FFont do begin
    Style := [];
    Charset := DEFAULT_CHARSET;
    Size := 8;
    Name := 'Tahoma';
  end;
  FTitleFont.Assign(FFont);
  FTitleFont.Size := 10;
  FTitleFont.Style := [fsBold];
  FTitleFont.Color := clGradientActiveCaption;
  FTopColor := $00FDFDFD;
  FBottomColor := clSilver;
end;

destructor TCustomOptions.Destroy;
begin
  if Assigned(FFont) then FFont.Free;
  if Assigned(FTitleFont) then FTitleFont.Free;
  if Assigned(FWallPaper) then FWallPaper.Free;
  inherited destroy;
end;

procedure TCustomOptions.SetWallPaper(Value: TBitmap);
begin
  if FWallPaper = nil then
    FWallPaper := TBitmap.Create;
  FWallPaper.Assign(Value);
end;

function TCustomOptions.GetWallPaper: TBitmap;
begin
  if FWallPaper = nil then
    FWallPaper := TBitmap.Create;
  Result := FWallPaper;
end;

///////////////////////////////  Input Options

constructor TInputOptions.Create(AOwner: TzBalloon);
begin
  inherited Create(AOwner);
  FEdit := TInputEdit.Create(AOwner);
  FEdit.FFont.Assign(FFont);
  FMin := 0;
  FMax := 0;
  FDecimalCount := 2;
  FButtonOptions := TButtonOptions.Create(AOwner);
  fBoxList := TStringList.Create;
  fComboList := TStringList.Create;
end;

destructor TInputOptions.Destroy;
begin
  fBoxList.Free;
  fComboList.Free;
end;

procedure TInputOptions.SetBoxList(Value: TStrings);
begin
  if fBoxList <> value then
  begin
    fBoxList.beginUpdate;
    fBoxList.Assign(Value);
    fBoxList.EndUpDate;
  end;
end;

procedure TInputOptions.SetComboList(Value: TStrings);
begin
  if fBoxList <> value then
  begin
    fBoxList.beginUpdate;
    fBoxList.Assign(Value);
    fBoxList.EndUpDate;
  end;
end;

///////////////////////////////  Dialog Options

constructor TDlgOptions.Create(AOwner: TzBalloon);
begin
  inherited Create(AOwner);
  FButtonOptions := TButtonOptions.Create(AOwner);
end;

///////////////////////////////  InputEdit

constructor TInputEdit.Create(AOwner: TzBalloon);
begin
  inherited Create;
  FColor := clWindow;
  FFont := TFont.Create;
//  FFont.Assign(AOwner.InputOptions.Font);
end;

destructor TInputEdit.Destroy;
begin
  if Assigned(FFont) then FFont.Free;
  inherited destroy;
end;

/////////////////////////////// zzButton

constructor TzzButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [csSetCaption, csCaptureMouse, csClickEvents	];
  Width := 75;
  Height := 25;
  TabStop := True;
  Cursor := crHandPoint;
end;

procedure TzzButton.CMFocusChanged(var Message: TCMFocusChanged);
begin
  with Message do
    if Sender is TzzButton then
      FActive := Sender = Self
    else
      FActive := False;
  inherited;
  Invalidate;
end;

procedure TzzButton.WMLBUTTONDOWN( Var msg: TWMLBUTTONDOWN);
begin
  inherited;
  FDown := True;
  Invalidate;
end;

procedure TzzButton.WMLBUTTONUP( Var msg: TWMLBUTTONUP);
begin
  inherited;
  FDown := False;
  Invalidate;
end;

procedure TzzButton.CMMOUSELEAVE(var Message: TMessage);
begin
  inherited;
  FMouseIn := False;
  Invalidate;
end;

procedure TzzButton.CMMOUSEENTER(var Message: TMessage);
begin
  inherited;
  FMouseIn := True;
  Invalidate;
end;

procedure TzzButton.Paint;
var
  FontHeight: Integer;
  Flags: Longint;
  Rect: TRect;
  x, y, v, h: Integer;
  PixColor: TColor;
begin
  Rect := GetClientRect;
  with Canvas do begin
    Brush.Style := bsClear;
    if Assigned(zBalloonControl.FButtonOptions.FWallPaper) then begin
      v := Height div zBalloonControl.FButtonOptions.FWallPaper.Height + 1;
      h := Width div zBalloonControl.FButtonOptions.FWallPaper.Width + 1;
      for x := 0 to h do
        for y := 0 to v do
          Draw(x * zBalloonControl.FButtonOptions.FWallPaper.Width,
               y * zBalloonControl.FButtonOptions.FWallPaper.Height,
               zBalloonControl.FButtonOptions.FWallPaper);
    end
    else if FDown then
      DrawGradient(Self.Handle,
                   zBalloonControl.FButtonOptions.FBottomColor,
                   zBalloonControl.FButtonOptions.FTopColor)
    else DrawGradient(Self.Handle,
                      zBalloonControl.FButtonOptions.FTopColor,
                      zBalloonControl.FButtonOptions.FBottomColor);

    Pen.Color := clGray;
    RoundRect(0, 0, Width, Height, 2, 2);

    if FMouseIn then begin
      if FDown then begin
        Pen.Color := clGray;
        MoveTo(1, 1);
        LineTo(Width - 1, 1);
        Pen.Color := clSilver;
        MoveTo(1, Height - 2);
        LineTo(Width - 1, Height - 2);
      end
      else begin
        Pen.Color := clSilver;
        MoveTo(1, 1);
        LineTo(Width - 1, 1);
        Pen.Color := $00FDFDFD;
        MoveTo(1, Height - 2);
        LineTo(Width - 1, Height - 2);
      end;
    end;

    if zBalloonControl.FShowMode = smInput then
      PixColor := zBalloonControl.FInputOptions.FBottomColor
    else PixColor := zBalloonControl.FDlgOptions.FBottomColor;
    Pixels[0, 0] := PixColor;
    Pixels[0, Rect.Bottom - 1] := PixColor;
    Pixels[Rect.Right - 1, 0] := PixColor;
    Pixels[Rect.Right - 1, Rect.Bottom - 1] := PixColor;

    Font := zBalloonControl.FButtonOptions.Font;
    if Focused then begin
      Font.Style := [fsBold];
    end;
    FontHeight := TextHeight('W');
    with Rect do
    begin
      Top := ((Bottom + Top) - FontHeight) div 2;
      Bottom := Top + FontHeight;
    end;
    Flags := DT_EXPANDTABS or DT_VCENTER or DT_CENTER;
    Flags := DrawTextBiDiModeFlags(Flags);
    if not Enabled then
    begin
      OffsetRect(Rect, 1, 1);
      Canvas.Font.Color := clBtnHighlight;
      DrawText(Canvas.Handle, PChar(Caption), Length(Caption), Rect, Flags);
      OffsetRect(Rect, -1, -1);
      Canvas.Font.Color := clBtnShadow;
      DrawText(Canvas.Handle, PChar(Caption), Length(Caption), Rect, Flags);
    end
    else DrawText(Handle, PChar(Caption), -1, Rect, Flags);
  end;
end;

procedure TzzButton.Click;
var
  Form: TCustomForm;
begin
  if not zBalloonControl.FSoundsOff then
    PlaySound(PChar('BUTTON_SOUND'), 0, SND_RESOURCE or SND_ASYNC);
  Form := GetParentForm(Self);
  if Form <> nil then Form.ModalResult := ModalResult;
  Repaint;
  inherited Click;
end;

procedure TzzButton.CMDialogKey(var Message: TCMDialogKey);
begin
  with Message do
    if  (((CharCode = VK_RETURN) and FActive) or
      ((CharCode = VK_ESCAPE) and FCancel)) and
      (KeyDataToShiftState(Message.KeyData) = []) and CanFocus then
    begin
      Click;
      Result := 1;
    end else
      inherited;
end;


/////////////////////////////////////////  Button Captions

constructor TButtonCaptions.Create(AOwner: TzBalloon);
begin
  inherited Create;
  FYes := 'Yes';
  FNo := 'No';
  FCancel := 'Cancel';
  FOk := 'Ok';
end;

///////////////////////////////////////// TzBalloonCheckBox


procedure TzBalloonCheckBox.CNDrawItem(var Message: TWMDrawItem);
var
  XCanvas: TCanvas;
  XCaptionRect, XGlyphRect: TRect;

  procedure xxDrawBitMap(ACanvas: TCanvas);
  const
    xx_h = 13;
    xx_w = 13;
  var
    xxGlyph: TBitmap;
    xxX, xxY, xxStepY, xxStepX: integer;
  begin
    xxGlyph := TBitmap.Create;
    try
      xxGlyph.Handle := LoadBitmap(0, PChar(OBM_CHECKBOXES));
      xxY := XGlyphRect.Top + (XGlyphRect.Bottom - XGlyphRect.Top - xx_h) div 2;
      xxX := 2;
      xxStepX := 0;
      xxStepY := 0;
      case State of
        cbChecked: xxStepX := xxStepX + xx_w;
        cbGrayed:  xxStepX := xxStepX + xx_w * 3;
        end;
      ACanvas.CopyRect(Rect(xxX, xxY, xxX + xx_w,  xxY + xx_h), xxGlyph.Canvas,
                       Rect(xxStepX, xxStepY, xx_w + xxStepX, xx_h + xxStepY));
    finally
      xxGlyph.Free;
      end;
    end;

  procedure xxDrawCaption;
  var
    xXFormat: longint;
  begin
    xXFormat := DT_VCENTER + DT_SINGLELINE + DT_LEFT;
    xXFormat := DrawTextBiDiModeFlags(xXFormat);
    DrawText(Message.DrawItemStruct.hDC, PChar(Caption), length(Caption),
                      XCaptionRect, xXFormat);
    end;

begin
  XGlyphRect := Message.DrawItemStruct.rcItem;
  XGlyphRect.Right := 20;
  XCaptionRect := Message.DrawItemStruct.rcItem;
  XCaptionRect.Left := XGlyphRect.Right;
  XCanvas := TCanvas.Create;
  try
    XCanvas.Handle := Message.DrawItemStruct.hDC;
    XCanvas.Brush.Style := bsClear;
    xxDrawBitMap(XCanvas);
    xxDrawCaption;
  finally
    XCanvas.Free;
    end;
  end;


procedure TzBalloonCheckBox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.ExStyle := Params.ExStyle or WS_EX_Transparent;
  end;


procedure TzBalloonCheckBox.CreateWnd;
begin
  inherited CreateWnd;
  SetButtonStyle;
  end;


procedure TzBalloonCheckBox.SetButtonStyle;
const
  BS_MASK = $000F;
var
  Style: Word;
begin
  if HandleAllocated then
  begin
    Style := BS_CHECKBOX or BS_OWNERDRAW;
    if GetWindowLong(Handle, GWL_STYLE) and BS_MASK <> Style then
      SendMessage(Handle, BM_SETSTYLE, Style, 1);
  end;
end;

/////////////////////////////////////////

procedure Register;
begin
  RegisterComponents('zPage', [TzBalloon]);
end;

end.
